/*
 * PostgreSQL functions to interface with memcache.
 *
 * Copyright (c) 2004-2005 Sean Chittenden <sean@chittenden.org>
 * Copyright (c) 2007 Neil Conway <neilc@samurai.com>
 * Copyright (c) 2007 Open Technology Group, Inc. <http://www.otg-nc.com>
 *
 * See the file COPYING for distribution terms.
 */

#include "postgres.h"

#include <inttypes.h>

#include "access/heapam.h"
#include "access/htup.h"
#include "fmgr.h"
#include "funcapi.h"
#include "lib/stringinfo.h"
#include "utils/builtins.h"
#include "utils/datetime.h"
#include "utils/guc.h"
#include "utils/memutils.h"

#include "pgmemcache.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/*
 * PostgreSQL <= 8.2 provided VARATT_SIZEP(p), but later versions use
 * SET_VARSIZE() (and also define the VARATT_SIZEP_DEPRECATED
 * symbol). Therefore, define the latter in terms of the former if
 * necessary.
 */
#ifndef SET_VARSIZE
#define SET_VARSIZE(ptr, size) VARATT_SIZEP(ptr) = (size)
#endif

/* Per-backend global state. */
struct memcache_global
{
    struct memcache *mc;
    struct memcache_ctxt *mc_ctxt;

    /* context in which long-lived state is allocated */
    MemoryContext pg_ctxt;
};

static struct memcache_global globals;

/* Custom GUC variable */
static char *memcache_default_servers;

static void      *mcm_palloc(const size_t size);
static const char *assign_default_servers_guc(const char *newval,
                                              bool doit, GucSource source);
static const char *show_default_servers_guc(void);
static Datum       memcache_atomic_op(bool increment, PG_FUNCTION_ARGS);
static Datum       memcache_set_cmd(int type, PG_FUNCTION_ARGS);
static bool        do_server_add(const char *host, size_t host_len,
                                 const char *port, size_t port_len);
static HeapTuple   form_server_tuple(struct memcache_server *serv,
                                     TupleDesc tdesc);
static bool        do_memcache_set_cmd(int type, char *key, size_t key_len,
                                       char *val, size_t val_len, time_t expire);
static time_t      interval_to_time_t(Interval *span);

#define MCM_SET_CMD_TYPE_ADD            0x0001
#define MCM_SET_CMD_TYPE_REPLACE        0x0002
#define MCM_SET_CMD_TYPE_SET            0x0004

#define MCM_DATE_TYPE_INTERVAL          0x0010
#define MCM_DATE_TYPE_TIMESTAMP         0x0020

/*
 * palloc() is a macro, so supply a version we can pass as a function
 * pointer.
 */
static void *
mcm_palloc(const size_t size)
{
    return palloc(size);
}

void
_PG_init(void)
{
    MemoryContext old_ctxt;

    globals.pg_ctxt = AllocSetContextCreate(TopMemoryContext,
                                            "pgmemcache global context",
                                            ALLOCSET_SMALL_MINSIZE,
                                            ALLOCSET_SMALL_INITSIZE,
                                            ALLOCSET_SMALL_MAXSIZE);

    old_ctxt = MemoryContextSwitchTo(globals.pg_ctxt);

    globals.mc_ctxt = mcMemNewCtxt(pfree, mcm_palloc, NULL, repalloc);
    if (globals.mc_ctxt == NULL)
        elog(ERROR, "failed to create libmemcache memory context");

    globals.mc = mcm_new(globals.mc_ctxt);
    if (globals.mc == NULL)
        elog(ERROR, "failed to create libmemcache instance");

    MemoryContextSwitchTo(old_ctxt);

    DefineCustomStringVariable("pgmemcache.default_servers",
                               "Comma-separated list of memcached servers to connect to.",
                               "Specified as a comma-separated list "
                               "of host:port (port is optional).",
                               &memcache_default_servers,
                               PGC_USERSET,
                               assign_default_servers_guc,
                               show_default_servers_guc);
}

static const char *
assign_default_servers_guc(const char *newval, bool doit, GucSource source)
{
    int            i;
    int            len;
    StringInfoData host_buf;
    StringInfoData port_buf;

    initStringInfo(&host_buf);
    initStringInfo(&port_buf);

    len = strlen(newval);

    for (i = 0; i < len; i++)
    {
        char   c = newval[i];

        /* The end of the hostname? */
        if (c == ',' || c == ':')
        {
            if (host_buf.len == 0)
                return NULL;

            /* Parse the (optional) port, if ':' found */
            if (c == ':')
            {
                int    j;

                for (j = i + 1; j < len; j++)
                {
                    if (newval[j] == ',')
                        break;

                    if (!isdigit((unsigned char) newval[j]))
                        return NULL;

                    appendStringInfoChar(&port_buf, newval[j]);
                }

                if (port_buf.len == 0)
                    return NULL;

                /* Skip the port itself, plus the port separator */
                i += port_buf.len;
            }

            if (doit)
                do_server_add(host_buf.data, host_buf.len,
                              port_buf.data, port_buf.len);

            /* Skip the element separator, reset buffers */
            i++;
            host_buf.data[0] = '\0';
            host_buf.len = 0;
            port_buf.data[0] = '\0';
            port_buf.len = 0;
        }
        else
        {
            /* Otherwise, assume it's part of the hostname */
            appendStringInfoChar(&host_buf, c);
        }
    }

    /* Last element of the list might not be terminated by a separator */
    if (host_buf.len > 0 && doit)
        do_server_add(host_buf.data, host_buf.len,
                      port_buf.data, port_buf.len);

    pfree(host_buf.data);
    pfree(port_buf.data);

    return newval;
}

static const char *
show_default_servers_guc(void)
{
    printf("%s called\n", __FUNCTION__);
    return memcache_default_servers;
}

/*
 * This is called when we're being unloaded from a process. Note that
 * this only happens when we're being replaced by a LOAD (e.g. it
 * doesn't happen on process exit), so we can't depend on it being
 * called.
 */
void
_PG_fini(void)
{
    mcm_free(globals.mc_ctxt, globals.mc);
    mcMemFreeCtxt(globals.mc_ctxt);
    MemoryContextDelete(globals.pg_ctxt);
}

Datum
memcache_add(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_ADD | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_add_absexpire(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_ADD | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


static Datum
memcache_atomic_op(bool increment, PG_FUNCTION_ARGS)
{
    text      *key = PG_GETARG_TEXT_P(0);
    size_t     key_len;
    u_int32_t  val;
    u_int32_t  ret;

    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        elog(ERROR, "unable to have a zero length key");

    val = 1;
    if (PG_NARGS() >= 2)
        val = PG_GETARG_UINT32(1);

    if (increment)
      ret = mcm_incr(globals.mc_ctxt, globals.mc, VARDATA(key), key_len, val);
    else
      ret = mcm_decr(globals.mc_ctxt, globals.mc, VARDATA(key), key_len, val);

    PG_RETURN_UINT32(ret);
}


Datum
memcache_decr(PG_FUNCTION_ARGS)
{
    return memcache_atomic_op(false, fcinfo);
}

Datum
memcache_delete(PG_FUNCTION_ARGS)
{
    text   *key = PG_GETARG_TEXT_P(0);
    size_t  key_len;
    time_t  hold;
    int     ret;

    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be the empty string")));

    hold = (time_t) 0.0;
    if (PG_NARGS() >= 2 && PG_ARGISNULL(1) == false)
        hold = interval_to_time_t(PG_GETARG_INTERVAL_P(1));

    ret = mcm_delete(globals.mc_ctxt, globals.mc,
                     VARDATA(key), key_len, hold);

    if (ret < 0)
        elog(ERROR, "internal libmemcache(3) error");

    PG_RETURN_BOOL(ret == 0);
}


static time_t
interval_to_time_t(Interval *span)
{
    float8 result;

#ifdef HAVE_INT64_TIMESTAMP
    result = span->time / 1000000e0;
#else
    result = span->time;
#endif

    if (span->month != 0)
    {
        result += (365.25 * 86400) * (span->month / 12);
        result += (30.0 * 86400) * (span->month % 12);
    }

    return (time_t) result;
}


Datum
memcache_flush_all0(PG_FUNCTION_ARGS)
{
    int     ret;

    ret = mcm_flush_all(globals.mc_ctxt, globals.mc);

    if (ret < 0)
        elog(ERROR, "internal libmemcache error");

    PG_RETURN_BOOL(ret == 0);
}


Datum
memcache_flush(PG_FUNCTION_ARGS)
{
    text                   *key = PG_GETARG_TEXT_P(0);
    size_t                  key_len;
    struct memcache_server *ms;
    int                     ret;
    u_int32_t               hash;

    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be the empty string")));

    hash = mcm_hash_key(globals.mc_ctxt, VARDATA(key), key_len);
    ms = mcm_server_find(globals.mc_ctxt, globals.mc, hash);

    ret = mcm_flush(globals.mc_ctxt, globals.mc, ms);

    if (ret < 0)
        elog(ERROR, "internal libmemcache error");

    PG_RETURN_BOOL(ret == 0);
}


Datum
memcache_get(PG_FUNCTION_ARGS)
{
    text                *key;
    text                *ret;
    size_t               key_len;
    struct memcache_req *req;
    struct memcache_res *res;

    if (PG_ARGISNULL(0))
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be NULL")));

    key = PG_GETARG_TEXT_P(0);
    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be the empty string")));

    req = mcm_req_new(globals.mc_ctxt);
    if (req == NULL)
        elog(ERROR, "unable to create a new memcache request structure");

    /*
     * Note that we pass the content of the key's data section directly
     * to libmemcache. We could call textout(), but we don't need a
     * NUL-terminated string, and textout() does a palloc(), which we'd
     * rather avoid.
     */
    res = mcm_req_add(globals.mc_ctxt, req, VARDATA(key), key_len);
    if (res == NULL)
        elog(ERROR, "unable to create a new memcache response structure");

    mcm_get(globals.mc_ctxt, globals.mc, req);

    if (!(res->_flags & MCM_RES_FOUND))
        PG_RETURN_NULL();

    ret = (text *) palloc(res->bytes + VARHDRSZ);
    SET_VARSIZE(ret, res->bytes + VARHDRSZ);
    memcpy(VARDATA(ret), res->val, res->bytes);

    PG_RETURN_TEXT_P(ret);
}


Datum
memcache_hash(PG_FUNCTION_ARGS)
{
    text      *key = PG_GETARG_TEXT_P(0);
    size_t     key_len;
    u_int32_t  hash;

    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be the empty string")));

    hash = mcm_hash_key(globals.mc_ctxt, VARDATA(key), key_len);
    PG_RETURN_UINT32(hash);
}


Datum
memcache_incr(PG_FUNCTION_ARGS)
{
    return memcache_atomic_op(true, fcinfo);
}


Datum
memcache_replace(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_REPLACE | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_replace_absexpire(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_REPLACE | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


Datum
memcache_set(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_SET | MCM_DATE_TYPE_INTERVAL, fcinfo);
}


Datum
memcache_set_absexpire(PG_FUNCTION_ARGS)
{
    return memcache_set_cmd(MCM_SET_CMD_TYPE_SET | MCM_DATE_TYPE_TIMESTAMP, fcinfo);
}


static Datum
memcache_set_cmd(int type, PG_FUNCTION_ARGS)
{
    text         *key;
    text         *val;
    size_t        key_len;
    size_t        val_len;
    time_t        expire;
    TimestampTz   timestamptz;
    struct pg_tm  tm;
    fsec_t        fsec;
    bool          ret;

    if (PG_ARGISNULL(0))
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be NULL")));
    if (PG_ARGISNULL(1))
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache value cannot be NULL")));

    key = PG_GETARG_TEXT_P(0);
    val = PG_GETARG_TEXT_P(1);

    val_len = VARSIZE(val) - VARHDRSZ;
    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache key cannot be the empty string")));

    expire = (time_t) 0.0;
    if (PG_NARGS() >= 3 && PG_ARGISNULL(2) == false)
    {
        if (type & MCM_DATE_TYPE_INTERVAL)
        {
            Interval *span = PG_GETARG_INTERVAL_P(2);
            expire = interval_to_time_t(span);
        }
        else if (type & MCM_DATE_TYPE_TIMESTAMP)
        {
            timestamptz = PG_GETARG_TIMESTAMPTZ(2);

            /* convert to timestamptz to produce consistent results */
            if (timestamp2tm(timestamptz, NULL, &tm, &fsec, NULL, NULL) !=0)
                ereport(ERROR,
                        (errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
                         errmsg("timestamp out of range")));

#ifdef HAVE_INT64_TIMESTAMP
            expire = (time_t) ((timestamptz - SetEpochTimestamp()) / 1000000e0);
#else
            expire = (time_t) timestamptz - SetEpochTimestamp();
#endif
        }
        else
        {
            elog(ERROR, "%s():%s:%u: invalid date type",
                 __FUNCTION__, __FILE__, __LINE__);
        }
    }

    ret = do_memcache_set_cmd(type, VARDATA(key), key_len,
                              VARDATA(val), val_len, expire);

    PG_RETURN_BOOL(ret);
}

static bool
do_memcache_set_cmd(int type, char *key, size_t key_len,
                    char *val, size_t val_len, time_t expire)
{
    int result = 0;

    if (type & MCM_SET_CMD_TYPE_ADD)
        result = mcm_add(globals.mc_ctxt, globals.mc, key, key_len, val, val_len, expire, 0);
    else if (type & MCM_SET_CMD_TYPE_REPLACE)
        result = mcm_replace(globals.mc_ctxt, globals.mc, key, key_len, val, val_len, expire, 0);
    else if (type & MCM_SET_CMD_TYPE_SET)
        result = mcm_set(globals.mc_ctxt, globals.mc, key, key_len, val, val_len, expire, 0);
    else
        elog(ERROR, "unknown memcache set command type: %d", type);

    if (result < 0)
        elog(ERROR, "internal libmemcache error");

    return (result == 0);
}


Datum
memcache_server_add(PG_FUNCTION_ARGS)
{
    text          *server = PG_GETARG_TEXT_P(0);
    char          *port_data;
    size_t         port_data_len;
    int            ret;

    /* If no port specified, use the libmemcache default (11211) */
    port_data = NULL;
    port_data_len = 0;
    if (PG_NARGS() > 1)
    {
        text *port = PG_GETARG_TEXT_P(1);
        port_data = VARDATA(port);
        port_data_len = VARSIZE(port) - VARHDRSZ;
    }

    ret = do_server_add(VARDATA(server), VARSIZE(server) - VARHDRSZ,
                        port_data, port_data_len);

    if (ret < 0)
    {
        elog(NOTICE, "unable to add memcache server: %d", ret);
        PG_RETURN_BOOL(false);
    }

    PG_RETURN_BOOL(true);
}

static bool
do_server_add(const char *host, size_t host_len,
              const char *port, size_t port_len)
{
    MemoryContext old_ctx;
    int           ret;

    /* Ensure that new servers are allocated in the long-term context */
    old_ctx = MemoryContextSwitchTo(globals.pg_ctxt);

    ret = mcm_server_add2(globals.mc_ctxt, globals.mc,
                          host, host_len, port, port_len);

    MemoryContextSwitchTo(old_ctx);

    return ret;
}

Datum
memcache_server_remove(PG_FUNCTION_ARGS)
{
    text                   *host = PG_GETARG_TEXT_P(0);
    char                   *host_str;
    char                   *port_str;
    struct memcache_server *serv;

    host_str = DatumGetCString(DirectFunctionCall1(textout,
                                                   PointerGetDatum(host)));

    /* If no port specified, use the libmemcache default (11211) */
    if (PG_NARGS() == 1)
        port_str = "11211";
    else
    {
        text *port = PG_GETARG_TEXT_P(1);

        port_str = DatumGetCString(DirectFunctionCall1(textout,
                                                       PointerGetDatum(port)));
    }

    /*
     * Find the first server that matches the specified host and port
     * in the server list, then remove it.
     */
    TAILQ_FOREACH(serv, &(globals.mc->server_list), entries)
    {
        if (strcmp(serv->hostname, host_str) == 0 &&
            strcmp(serv->port, port_str) == 0)
        {
            /*
             * XXX: all we do is disconnect from the server for
             * now. We should probably remove it from the servers list
             * as well, or rename this function.
             */
            mcm_server_disconnect(globals.mc_ctxt, serv);
            PG_RETURN_BOOL(true);
        }
    }

    PG_RETURN_BOOL(false);
}


static Datum
form_simple_server_tup(struct memcache_server *server, TupleDesc tdesc)
{
    Datum                   values[2];
    bool                    nulls[2];

    values[0] = DirectFunctionCall1(textin,
                                    CStringGetDatum(server->hostname));
    values[1] = DirectFunctionCall1(int4in,
                                    CStringGetDatum(server->port));
    nulls[0] = nulls[1] = false;

    return HeapTupleGetDatum(heap_form_tuple(tdesc, values, nulls));
}


Datum
memcache_server_find(PG_FUNCTION_ARGS)
{
    text                   *key = PG_GETARG_TEXT_P(0);
    size_t                  key_len;
    u_int32_t               hash;
    struct memcache_server *server;
    TupleDesc               tdesc;
    TypeFuncClass           func_type;

    key_len = VARSIZE(key) - VARHDRSZ;
    if (key_len < 1)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache server key cannot be the empty string")));

    hash = mcm_hash_key(globals.mc_ctxt, VARDATA(key), key_len);
    server = mcm_server_find(globals.mc_ctxt, globals.mc, hash);
    if (server == NULL)
      elog(ERROR, "failed to find memcache server for key");

    /*
     * XXX: the return type of this function is actually invariant, so
     * it seems a bit unfortunate to invoke get_call_result_type() for
     * every call, which can be expensive. Can we just hardcode the
     * expected tdesc?
     */
    func_type = get_call_result_type(fcinfo, NULL, &tdesc);
    if (func_type != TYPEFUNC_COMPOSITE)
      elog(ERROR, "called in unexpected context");

    PG_RETURN_DATUM(form_simple_server_tup(server, tdesc));
}


Datum
memcache_server_find_hash(PG_FUNCTION_ARGS)
{
    uint32                  hash = PG_GETARG_UINT32(0);
    struct memcache_server *server;
    TypeFuncClass           func_type;
    TupleDesc               tdesc;

    server = mcm_server_find(globals.mc_ctxt, globals.mc, hash);
    if (server == NULL)
      elog(ERROR, "failed to find memcache server for key");

    func_type = get_call_result_type(fcinfo, NULL, &tdesc);
    if (func_type != TYPEFUNC_COMPOSITE)
      elog(ERROR, "called in unexpected context");

    PG_RETURN_DATUM(form_simple_server_tup(server, tdesc));
}

Datum
memcache_server_list(PG_FUNCTION_ARGS)
{
    FuncCallContext *func_ctx;

    if (SRF_IS_FIRSTCALL())
    {
        MemoryContext old_ctx;
        TypeFuncClass func_type;
        TupleDesc     result_tdesc;

        func_ctx = SRF_FIRSTCALL_INIT();

        old_ctx = MemoryContextSwitchTo(func_ctx->multi_call_memory_ctx);

        func_type = get_call_result_type(fcinfo, NULL, &result_tdesc);
        if (func_type != TYPEFUNC_COMPOSITE)
            elog(ERROR, "called in unexpected context");

        func_ctx->tuple_desc = BlessTupleDesc(result_tdesc);
        func_ctx->max_calls = globals.mc->num_servers;

        MemoryContextSwitchTo(old_ctx);
    }

    func_ctx = SRF_PERCALL_SETUP();

    if (func_ctx->call_cntr < func_ctx->max_calls)
    {
        struct memcache_server *serv;
        HeapTuple               result;

        serv = globals.mc->servers[func_ctx->call_cntr];
        result = form_server_tuple(serv, func_ctx->tuple_desc);
        SRF_RETURN_NEXT(func_ctx, HeapTupleGetDatum(result));
    }

    SRF_RETURN_DONE(func_ctx);
}

static HeapTuple
form_server_tuple(struct memcache_server *serv, TupleDesc tdesc)
{
    Datum values[3];
    bool  nulls[3];
    char *state_str;

    values[0] = DirectFunctionCall1(textin, CStringGetDatum(serv->hostname));
    values[1] = DirectFunctionCall1(int4in, CStringGetDatum(serv->port));

    switch (serv->active)
    {
        case 'd':
            state_str = "Down";
            break;

        case 'n':
            state_str = "Unknown host";
            break;

        case 't':
            state_str = "Unconnected";
            break;

        case 'u':
            state_str = "Connected";
            break;

        default:
            elog(ERROR, "unrecognized server state: %c", serv->active);
            state_str = NULL;   /* keep compiler quiet */
    }

    values[2] = DirectFunctionCall1(textin, CStringGetDatum(state_str));

    nulls[0] = nulls[1] = nulls[2] = false;

    return heap_form_tuple(tdesc, values, nulls);
}

Datum
memcache_stats(PG_FUNCTION_ARGS)
{
    struct memcache_server_stats *stats;
    StringInfoData                buf;

    stats = mcm_stats(globals.mc_ctxt, globals.mc);
    if (stats == NULL)
        elog(ERROR, "unable to create a new memcache stats structure");

    initStringInfo(&buf);
    appendStringInfo(&buf, "pid: %u\n", stats->pid);
    appendStringInfo(&buf, "uptime: %lu\n", (unsigned long) stats->uptime);
    appendStringInfo(&buf, "time: %lu\n", (unsigned long) stats->time);
    appendStringInfo(&buf, "version: %s\n", stats->version);
    appendStringInfo(&buf, "rusage_user: %lu.%ld\n",
                     (unsigned long) stats->rusage_user.tv_sec,
                     (long) stats->rusage_user.tv_usec);
    appendStringInfo(&buf, "rusage_system: %lu.%ld\n",
                     (unsigned long) stats->rusage_system.tv_sec,
                     (long) stats->rusage_system.tv_usec);
    appendStringInfo(&buf, "curr_items: %u\n", stats->curr_items);
    appendStringInfo(&buf, "total_items: %" PRIu64 "\n", stats->total_items);
    appendStringInfo(&buf, "bytes: %" PRIu64 "\n", stats->bytes);
    appendStringInfo(&buf, "curr_connections: %u\n", stats->curr_connections);
    appendStringInfo(&buf, "total_connections: %" PRIu64 "\n", stats->total_connections);
    appendStringInfo(&buf, "connection_structures: %u\n", stats->connection_structures);
    appendStringInfo(&buf, "cmd_get: %" PRIu64 "\n", stats->cmd_get);
    appendStringInfo(&buf, "cmd_set: %" PRIu64 "\n", stats->cmd_set);
    appendStringInfo(&buf, "get_hits: %" PRIu64 "\n", stats->get_hits);
    appendStringInfo(&buf, "get_misses: %" PRIu64 "\n", stats->get_misses);
    appendStringInfo(&buf, "bytes_read: %" PRIu64 "\n", stats->bytes_read);
    appendStringInfo(&buf, "bytes_written: %" PRIu64 "\n", stats->bytes_written);
    appendStringInfo(&buf, "limit_maxbytes: %" PRIu64 "\n", stats->limit_maxbytes);

    PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(buf.data)));
}


Datum
memcache_stat(PG_FUNCTION_ARGS)
{
    text                         *stat = PG_GETARG_TEXT_P(0);
    char                         *stat_str;
    size_t                        stat_len;
    struct memcache_server_stats *stats;
    StringInfoData                buf;

    stat_str = DatumGetCString(DirectFunctionCall1(textout,
                                                   PointerGetDatum(stat)));
    stat_len = strlen(stat_str);
    if (stat_len == 0)
        ereport(ERROR,
                (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg("memcache statistic key cannot be the empty string")));

    stats = mcm_stats(globals.mc_ctxt, globals.mc);
    if (stats == NULL)
        elog(ERROR, "unable to create a new memcache stats structure");

    initStringInfo(&buf);

    if (pg_strncasecmp("pid", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%u", stats->pid);
    else if (pg_strncasecmp("uptime", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%lu", (unsigned long) stats->uptime);
    else if (pg_strncasecmp("time", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%lu", (unsigned long) stats->time);
    else if (pg_strncasecmp("version", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%s", stats->version);
    else if (pg_strncasecmp("rusage_user", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%lu.%ld",
                         (unsigned long) stats->rusage_user.tv_sec,
                         (long) stats->rusage_user.tv_usec);
    else if (pg_strncasecmp("rusage_system", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%lu.%ld",
                         (unsigned long) stats->rusage_system.tv_sec,
                         (long) stats->rusage_system.tv_usec);
    else if (pg_strncasecmp("curr_items", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%u", stats->curr_items);
    else if (pg_strncasecmp("total_items", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->total_items);
    else if (pg_strncasecmp("bytes", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->bytes);
    else if (pg_strncasecmp("curr_connections", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%u", stats->curr_connections);
    else if (pg_strncasecmp("total_connections", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->total_connections);
    else if (pg_strncasecmp("connection_structures", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%u", stats->connection_structures);
    else if (pg_strncasecmp("cmd_get", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->cmd_get);
    else if (pg_strncasecmp("cmd_set", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->cmd_set);
    else if (pg_strncasecmp("get_hits", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->get_hits);
    else if (pg_strncasecmp("get_misses", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->get_misses);
    else if (pg_strncasecmp("bytes_read", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->bytes_read);
    else if (pg_strncasecmp("bytes_written", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->bytes_written);
    else if (pg_strncasecmp("limit_maxbytes", stat_str, stat_len) == 0)
        appendStringInfo(&buf, "%" PRIu64, stats->limit_maxbytes);
    else
        elog(ERROR, "unknown memcache statistic \"%s\"", stat_str);

    PG_RETURN_DATUM(DirectFunctionCall1(textin, CStringGetDatum(buf.data)));
}
